@@ -1,4 +1,3 @@
-#!/usr/bin/env python
 # -*- Mode: Python -*-
 # GObject-Introspection - a framework for introspecting GObject libraries
 # Copyright (C) 2008-2010 Johan Dahlin
@@ -24,12 +23,14 @@ import errno
 import optparse
 import os
 import shutil
-import subprocess
+import stat
 import sys
 import tempfile
 import platform
+import shlex
 
-from giscanner import message
+import giscanner
+from giscanner import message, pkgconfig
 from giscanner.annotationparser import GtkDocCommentBlockParser
 from giscanner.ast import Include, Namespace
 from giscanner.dumper import compile_introspection_binary
@@ -95,7 +96,8 @@ def get_windows_option_group(parser):
 
 
 def _get_option_parser():
-    parser = optparse.OptionParser('%prog [options] sources')
+    parser = optparse.OptionParser('%prog [options] sources',
+                                   version='%prog ' + giscanner.__version__)
     parser.add_option('', "--quiet",
                       action="store_true", dest="quiet",
                       default=False,
@@ -134,6 +136,9 @@ def _get_option_parser():
     parser.add_option("-l", "--library",
                       action="append", dest="libraries", default=[],
                       help="libraries of this unit")
+    parser.add_option("", "--extra-library",
+                      action="append", dest="extra_libraries", default=[],
+                      help="Extra libraries to link the binary against")
     parser.add_option("-L", "--library-path",
                       action="append", dest="library_paths", default=[],
                       help="directories to search for libraries")
@@ -164,6 +169,11 @@ the latter is not specified.""")
     parser.add_option("", "--symbol-prefix",
                       action="append", dest="symbol_prefixes", default=[],
                       help="Remove this prefix from C symbols (function names)")
+    parser.add_option("", "--symbol-filter-cmd",
+                      action="store", dest="symbol_filter_cmd", default='',
+                      help='Filter symbols (function names) through the given '
+                           'shell command which will receive the symbol name as input '
+                           'to stdin and is expected to output the filtered results to stdout.')
     parser.add_option("", "--accept-unprefixed",
                       action="store_true", dest="accept_unprefixed", default=False,
                       help="""If specified, accept symbols and identifiers that do not
@@ -232,6 +242,11 @@ match the namespace prefix.""")
     parser.add_option("", "--include-last-in-src",
                       action="append", dest="include_last_src", default=[],
                       help="Header to include after the other headers in generated sources")
+    parser.add_option("", "--sources-top-dirs", default=[], action='append',
+                      help="Paths to the sources directories used to determine"
+                      " relative files locations to be used in the gir file."
+                      " This is especially useful when build dir and source dir are different"
+                      " and mirrored.")
 
     return parser
 
@@ -245,7 +260,7 @@ def passthrough_gir(path, f):
     parser.parse(path)
 
     writer = GIRWriter(parser.get_namespace())
-    f.write(writer.get_xml())
+    f.write(writer.get_encoded_xml())
 
 
 def test_codegen(optstring,
@@ -266,12 +281,12 @@ def test_codegen(optstring,
                                       include_last_src)
         gen.write()
     else:
-        _error("Invaild namespace %r" % (namespace, ))
+        _error("Invaild namespace '%s'" % (namespace, ))
     return 0
 
 
 def process_options(output, allowed_flags):
-    for option in output.split():
+    for option in output:
         for flag in allowed_flags:
             if not option.startswith(flag):
                 continue
@@ -280,18 +295,11 @@ def process_options(output, allowed_flags):
 
 
 def process_packages(options, packages):
-    args = [os.environ.get('PKG_CONFIG', 'pkg-config'), '--cflags']
-    args.extend(packages)
-    output = subprocess.Popen(args,
-                              stdout=subprocess.PIPE).communicate()[0]
-    if output is None:
-        # the error output should have already appeared on our stderr,
-        # so we just exit
-        return 1
+    flags = pkgconfig.cflags(packages)
     # Some pkg-config files on Windows have options we don't understand,
     # so we explicitly filter to only the ones we need.
     options_whitelist = ['-I', '-D', '-U', '-l', '-L']
-    filtered_output = list(process_options(output, options_whitelist))
+    filtered_output = list(process_options(flags, options_whitelist))
     parser = _get_option_parser()
     pkg_options, unused = parser.parse_args(filtered_output)
     options.cpp_includes.extend([os.path.realpath(f) for f in pkg_options.cpp_includes])
@@ -317,8 +325,8 @@ def extract_filelist(options):
     filenames = []
     if not os.path.exists(options.filelist):
         _error('%s: no such filelist file' % (options.filelist, ))
-    filelist_file = open(options.filelist, "r")
-    lines = filelist_file.readlines()
+    with open(options.filelist, "r") as filelist_file:
+        lines = filelist_file.readlines()
     for line in lines:
         # We don't support real C++ parsing yet, but we should be able
         # to understand C API implemented in C++ files.
@@ -337,8 +345,8 @@ def extract_filelist(options):
 
 def create_namespace(options):
     if options.strip_prefix:
-        print """g-ir-scanner: warning: Option --strip-prefix has been deprecated;
-see --identifier-prefix and --symbol-prefix."""
+        print("""g-ir-scanner: warning: Option --strip-prefix has been deprecated;
+see --identifier-prefix and --symbol-prefix.""")
         options.identifier_prefixes.append(options.strip_prefix)
 
     # We do this dance because the empty list has different semantics from
@@ -365,21 +373,24 @@ see --identifier-prefix and --symbol-prefix."""
 
 
 def create_transformer(namespace, options):
+    identifier_filter_cmd = shlex.split(options.identifier_filter_cmd)
+    symbol_filter_cmd = shlex.split(options.symbol_filter_cmd)
     transformer = Transformer(namespace,
                               accept_unprefixed=options.accept_unprefixed,
-                              identifier_filter_cmd=options.identifier_filter_cmd)
+                              identifier_filter_cmd=identifier_filter_cmd,
+                              symbol_filter_cmd=symbol_filter_cmd)
     transformer.set_include_paths(options.include_paths)
-    if options.passthrough_gir:
+    if options.passthrough_gir or options.reparse_validate_gir:
         transformer.disable_cache()
         transformer.set_passthrough_mode()
 
     for include in options.includes:
         if os.sep in include:
-            _error("Invalid include path %r" % (include, ))
+            _error("Invalid include path '%s'" % (include, ))
         try:
             include_obj = Include.from_string(include)
-        except:
-            _error("Malformed include %r\n" % (include, ))
+        except Exception:
+            _error("Malformed include '%s'\n" % (include, ))
         transformer.register_include(include_obj)
     for include_path in options.includes_uninstalled:
         transformer.register_include_uninstalled(include_path)
@@ -416,6 +427,7 @@ def create_source_scanner(options, args):
         filenames = extract_filelist(options)
     else:
         filenames = extract_filenames(args)
+    filenames = [os.path.realpath(f) for f in filenames]
 
     if platform.system() == 'Darwin':
         options.cpp_undefines.append('__BLOCKS__')
@@ -427,26 +439,36 @@ def create_source_scanner(options, args):
                        options.cpp_defines,
                        options.cpp_undefines,
                        cflags=options.cflags)
-    ss.parse_files(filenames)
-    ss.parse_macros(filenames)
-    return ss
+    try:
+        ss.parse_files(filenames)
+        ss.parse_macros(filenames)
+    finally:
+        for error in ss.get_errors():
+            print(error, file=sys.stderr)
+    return ss, filenames
 
 
 def write_output(data, options):
+    """Write encoded XML 'data' to the filename specified in 'options'."""
     if options.output == "-":
         output = sys.stdout
     elif options.reparse_validate_gir:
         main_f, main_f_name = tempfile.mkstemp(suffix='.gir')
-        main_f = os.fdopen(main_f, 'w')
-        main_f.write(data)
-        main_f.close()
+
+        if (os.path.isfile(options.output)):
+            shutil.copystat(options.output, main_f_name)
+        else:
+            os.chmod(main_f_name,
+                     stat.S_IWUSR | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
+
+        with os.fdopen(main_f, 'wb') as main_f:
+            main_f.write(data)
 
         temp_f, temp_f_name = tempfile.mkstemp(suffix='.gir')
-        temp_f = os.fdopen(temp_f, 'w')
-        passthrough_gir(main_f_name, temp_f)
-        temp_f.close()
+        with os.fdopen(temp_f, 'wb') as temp_f:
+            passthrough_gir(main_f_name, temp_f)
         if not utils.files_are_identical(main_f_name, temp_f_name):
-            _error("Failed to re-parse gir file; scanned=%r passthrough=%r" % (
+            _error("Failed to re-parse gir file; scanned='%s' passthrough='%s'" % (
                 main_f_name, temp_f_name))
         os.unlink(temp_f_name)
         try:
@@ -454,12 +476,11 @@ def write_output(data, options):
         except OSError as e:
             if e.errno == errno.EPERM:
                 os.unlink(main_f_name)
-                return 0
             raise
         return 0
     else:
         try:
-            output = open(options.output, "w")
+            output = open(options.output, 'wb')
         except IOError as e:
             _error("opening output for writing: %s" % (e.strerror, ))
 
@@ -469,10 +490,41 @@ def write_output(data, options):
         _error("while writing output: %s" % (e.strerror, ))
 
 
+def get_source_root_dirs(options, filenames):
+    if options.sources_top_dirs:
+        return [os.path.realpath(p) for p in options.sources_top_dirs]
+
+    # None passed, we need to guess
+    filenames = [os.path.realpath(p) for p in filenames]
+    dirs = sorted(set([os.path.dirname(f) for f in filenames]))
+
+    # We need commonpath (3.5+), otherwise give up
+    if not hasattr(os.path, "commonpath"):
+        return dirs
+
+    if not dirs:
+        return []
+
+    try:
+        common = os.path.commonpath(dirs)
+    except ValueError:
+        # ValueError: On Windows in case the paths are on different drives
+        return dirs
+
+    # If the only common path is the root directory give up
+    if os.path.dirname(common) == common:
+        return dirs
+
+    return [common]
+
+
 def scanner_main(args):
     parser = _get_option_parser()
     (options, args) = parser.parse_args(args)
 
+    if options.verbose:
+        import distutils
+        distutils.log.set_threshold(distutils.log.DEBUG)
     if options.passthrough_gir:
         passthrough_gir(options.passthrough_gir, sys.stdout)
     if options.test_codegen:
@@ -503,18 +555,19 @@ def scanner_main(args):
     namespace = create_namespace(options)
     logger = message.MessageLogger.get(namespace=namespace)
     if options.warn_all:
-        logger.enable_warnings((message.WARNING, message.ERROR, message.FATAL))
+        logger.enable_warnings(True)
 
     transformer = create_transformer(namespace, options)
 
     packages = set(options.packages)
     packages.update(transformer.get_pkgconfig_packages())
     if packages:
-        exit_code = process_packages(options, packages)
-        if exit_code:
-            return exit_code
+        try:
+            process_packages(options, packages)
+        except pkgconfig.PkgConfigError as e:
+            _error(str(e))
 
-    ss = create_source_scanner(options, args)
+    ss, filenames = create_source_scanner(options, args)
 
     cbp = GtkDocCommentBlockParser()
     blocks = cbp.parse_comment_blocks(ss.get_comments())
@@ -541,9 +594,10 @@ def scanner_main(args):
     if options.warn_fatal and warning_count > 0:
         message.fatal("warnings configured as fatal")
         return 1
-    elif warning_count > 0 and options.warn_all is False:
-        print ("g-ir-scanner: %s: warning: %d warnings suppressed (use --warn-all to see them)"
-               % (transformer.namespace.name, warning_count, ))
+    elif warning_count > 0 and options.warn_all is False and options.quiet is False:
+        print("g-ir-scanner: %s: warning: %d warnings suppressed "
+              "(use --warn-all to see them)" %
+              (transformer.namespace.name, warning_count, ))
 
     # Write out AST
     if options.packages_export:
@@ -553,8 +607,10 @@ def scanner_main(args):
 
     transformer.namespace.c_includes = options.c_includes
     transformer.namespace.exported_packages = exported_packages
-    writer = Writer(transformer.namespace)
-    data = writer.get_xml()
+
+    sources_top_dirs = get_source_root_dirs(options, filenames)
+    writer = Writer(transformer.namespace, sources_top_dirs)
+    data = writer.get_encoded_xml()
 
     write_output(data, options)
 
