@@ -1,4 +1,3 @@
-#!/usr/bin/env python
 # -*- Mode: Python -*-
 # GObject-Introspection - a framework for introspecting GObject libraries
 # Copyright (C) 2009 Red Hat, Inc.
@@ -19,12 +18,8 @@
 # 02110-1301, USA.
 #
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-from __future__ import unicode_literals
-
 import os
+import sys
 import platform
 import re
 import subprocess
@@ -106,18 +101,35 @@ def _resolve_non_libtool(options, binary, libraries):
             args.extend(['otool', '-L', binary.args[0]])
         else:
             args.extend(['ldd', binary.args[0]])
-        output = subprocess.check_output(args, universal_newlines=True, errors='replace')
+        output = subprocess.check_output(args)
+        if isinstance(output, bytes):
+            output = output.decode("utf-8", "replace")
 
-        # Use absolute paths on OS X to conform to how libraries are usually
-        # referenced on OS X systems, and file names everywhere else.
-        basename = platform.system() != 'Darwin'
-        return resolve_from_ldd_output(libraries, output, basename=basename)
+        shlibs = resolve_from_ldd_output(libraries, output)
+        return list(map(sanitize_shlib_path, shlibs))
 
 
-def resolve_from_ldd_output(libraries, output, basename=False):
+def sanitize_shlib_path(lib):
+    # Use absolute paths on OS X to conform to how libraries are usually
+    # referenced on OS X systems, and file names everywhere else.
+    # In case we get relative paths on macOS (like @rpath) then we fall
+    # back to the basename as well:
+    # https://gitlab.gnome.org/GNOME/gobject-introspection/issues/222
+    if sys.platform == "darwin":
+        if not os.path.isabs(lib):
+            return os.path.basename(lib)
+        return lib
+    else:
+        return os.path.basename(lib)
+
+
+def resolve_from_ldd_output(libraries, output):
     patterns = {}
     for library in libraries:
-        patterns[library] = _ldd_library_pattern(library)
+        if not os.path.isfile(library):
+            patterns[library] = _ldd_library_pattern(library)
+    if len(patterns) == 0:
+        return []
 
     shlibs = []
     for line in output.splitlines():
@@ -139,8 +151,6 @@ def resolve_from_ldd_output(libraries, output, basename=False):
             "ERROR: can't resolve libraries to shared libraries: " +
             ", ".join(patterns.keys()))
 
-    if basename:
-        shlibs = list(map(os.path.basename, shlibs))
     return shlibs
 
 
